#ifndef __KDATA_DOT_H__
#define __KDATA_DOT_H__

#include <stdbool.h>

extern struct u_block ub;
extern struct u_data udata;

#define BOOT_TTY 37       /* Set this to default device for stdio, stderr */
                          /* In this case, the default is the first TTY device */

#define NUM_DEV_TTY 2
#define NUM_DEV_SD 28
#define NUM_DEV_RD 4
#define NDEVS    36       /* Devices 0..NDEVS-1 are capable of being mounted */
                          /*  (add new mountable devices to beginning area.) */
#define TTYDEV   BOOT_TTY /* Device used by kernel for messages, panics */
/* #define SWAPDEV  10 */       /* Device for swapping. */
#define NBUFS    10       /* Number of block buffers */

extern uint16_t ramsize, procmem, maxproc;
extern uint16_t system_tick_counter;

extern struct p_tab ptab[PTABSIZE];
extern unsigned maxproc;   /* Actual max number of processes */
extern unsigned ramsize;
extern unsigned procmem;

extern inoptr root;        /* Address of root dir in inode table */
extern uint8_t root_dev;   /* Device number of root filesystem. */

extern struct cinode i_tab[ITABSIZE];    /* In-core inode table */
extern struct oft of_tab[OFTSIZE];       /* Open File Table */

extern struct filesys fs_tab[NDEVS];
extern struct blkbuf bufpool[NBUFS];

extern ptptr initproc;  /* The process table address of the first process. */
extern bool inint;     /* flag is set whenever interrupts are being serviced */

extern uint8_t ticks_this_second;       /* Tick counter for counting off one second */
extern uint16_t runticks;  /* Number of ticks current process has been swapped in */

extern uzitime_t tod;      /* Time of day */
extern uzitime_t ticks;    /* Cumulative tick counter, in minutes and ticks  */

extern char *swapbase;  /* Used by device driver for swapping */
extern unsigned swapcnt;
extern blkno_t swapblk;

extern uint16_t waitno;   /* Serial number of processes entering wait state */

extern struct devsw dev_tab[];

// The device driver switch table
typedef int (*dev_read_t)(uint8_t minor, uint8_t rawflag);
typedef int (*dev_write_t)(uint8_t minor, uint8_t rawflag);
typedef int (*dev_init_t)(void);
typedef int (*dev_open_t)(uint8_t minor);
typedef int (*dev_close_t)(uint8_t minor);
typedef int (*dev_ioctl_t)(uint8_t minor, uint16_t request, char *data); // note: data is in userspace

typedef struct devsw {
    int minor;          /* The minor device number (an argument to below) */
    dev_open_t dev_open;  /* The routines for reading, etc */
    dev_close_t dev_close; /* Format: op(minor,blkno,offset,count,buf); */
    dev_read_t dev_read;  /* Offset would be ignored for block devices */
    dev_write_t dev_write; /* Blkno and offset ignored for tty, etc. */
    dev_ioctl_t dev_ioctl; /* Count is rounded to 512 for block devices */
} devsw;

// the system call dispatch table
#define UZI_SYSCALL_COUNT 46
typedef int16_t (*syscall_t)(void);
extern syscall_t syscall_dispatch[UZI_SYSCALL_COUNT];

#endif
