/****************************************************
UZI (Unix Z80 Implementation) Kernel:  unix.h
From UZI by Doug Braun and UZI280 by Stefan Nitschke.
*****************************************************/
/* History:
 *   21.12.97 - Removed leading ? from Per-Process equates.	HFB
 *   11.07.98 - Shortened Time Slices to 50 Ticks/Sec (20 mS).	HFB
 */

#ifndef __UZI__KERNEL_DOT_H__
#define __UZI__KERNEL_DOT_H__

// #define __critical if(1)

#include <stdbool.h>

#define time_t DO_NOT_USE_STANDARD_time_t_USE_uzitime_t_INSTEAD

#ifndef NULL
#define NULL (void *)0
#endif

#define min(a,b) ( (a) < (b) ? (a) : (b) )
#define max(a,b) ( (a) > (b) ? (a) : (b) )

#define CPM_EMULATOR_FILENAME    "/usr/cpm/emulator"

#define UFTSIZE 10       /* Number of user files */		/*280 = 22*/
#define OFTSIZE 15       /* Open file table size */		/*280 = 45*/
#define ITABSIZE 20      /* Inode table size */			/*280 = 45*/
#define PTABSIZE 15      /* Process table size. */

#define TICKSPERSEC 100   /* Ticks per second */
#define MAXTICKS      1   /* Max ticks before swapping out (time slice)
                            default process time slice */
// #define MAXBACK      3   /* Process time slice for tasks not connected
//                             to the current tty */
// #define MAXBACK2     2   /* Process time slice for background tasks */
// #define MAXINTER     5   /* Process time slice for interactive tasks */

#define MAXPID    32000

#define PROGBASE    ((char *)(0x0100))  /* also data base */
#define PROGTOP     ((char *)(0xf900))  /* Top of program, base of U_DATA */

#define NSIGS 16         /* Number of signals <= 16 */
#define ROOTINODE 1      /* Inode # of / for all mounted filesystems. */
#define EMAGIC    0xc3   /* Header of executable */
#define CMAGIC   24721   /* Random number for cinode c_magic */
#define SMOUNTED 12742   /* Magic number to specify mounted filesystem */

#define OS_BANK 0
#define NO_DEVICE (0xFF)
#define NO_FILE   (0xFF)

typedef unsigned long uint32_t;
typedef signed long int32_t;
typedef unsigned short uint16_t;
typedef signed short int16_t;
typedef unsigned char uint8_t;
typedef signed char int8_t;

typedef struct s_queue {
    char *q_base;    /* Pointer to data */
    char *q_head;    /* Pointer to addr of next char to read. */
    char *q_tail;    /* Pointer to where next char to insert goes. */
    int   q_size;    /* Max size of queue */
    int   q_count;   /* How many characters presently in queue */
    int   q_wakeup;  /* Threshold for waking up processes waiting on queue */
} queue_t;

typedef struct uzitime_s {
    uint16_t t_time;
    uint16_t t_date;
} uzitime_t;

/* User's structure for times() system call */
struct tms {
	uzitime_t  tms_utime;
	uzitime_t  tms_stime;
	uzitime_t  tms_cutime;
	uzitime_t  tms_cstime;
	uzitime_t  tms_etime;      /* Elapsed real time */
} ;

/* Flags for setftime() */
#define A_TIME 1
#define M_TIME 2
#define C_TIME 4

typedef struct off_t {
    uint16_t o_blkno;      /* Block number */
    int16_t  o_offset;     /* Offset within block 0-511 */
} off_t;

typedef uint16_t blkno_t;    /* Can have 65536 512-byte blocks in filesystem */
#define NULLBLK ((blkno_t)-1)

typedef struct blkbuf {
    char        bf_data[512];    /* This MUST be first ! */
    uint8_t     bf_dev;
    blkno_t     bf_blk;
    bool        bf_dirty;
    bool        bf_busy;
    uint16_t    bf_time;         /* LRU time stamp */
} blkbuf, *bufptr;

typedef struct dinode {
    uint16_t i_mode;
    uint16_t i_nlink;
    uint16_t i_uid;
    uint16_t i_gid;
    off_t    i_size;
    uzitime_t   i_atime;
    uzitime_t   i_mtime;
    uzitime_t   i_ctime;
    blkno_t  i_addr[20];
} dinode;               /* Exactly 64 bytes long! */

struct  stat    /* Really only used by users */
{
	int16_t   st_dev;
	uint16_t  st_ino;
	uint16_t  st_mode;
	uint16_t  st_nlink;
	uint16_t  st_uid;
	uint16_t  st_gid;
	uint16_t  st_rdev;
	off_t   st_size;
	uzitime_t  st_atime;
	uzitime_t  st_mtime;
	uzitime_t  st_ctime;
};

/* Bit masks for i_mode and st_mode */

#define OTH_EX  0001
#define OTH_WR  0002
#define OTH_RD  0004
#define GRP_EX  0010
#define GRP_WR  0020
#define GRP_RD  0040
#define OWN_EX  0100
#define OWN_WR  0200
#define OWN_RD  0400

#define SAV_TXT 01000
#define SET_GID 02000
#define SET_UID 04000

#define MODE_MASK 07777

#define F_REG   0100000
#define F_DIR   040000
#define F_PIPE  010000
#define F_BDEV  060000  // important that F_BDEV & F_CDEV != 0 (see isdevice() function)
#define F_CDEV  020000

#define F_MASK  0170000

typedef struct cinode { // note: exists in memory *and* on disk
    uint16_t   c_magic;           /* Used to check for corruption. */
    uint8_t    c_dev;             /* Inode's device */
    uint8_t    c_dev_highbyte;    /* c_dev future expansion */
    uint16_t   c_num;             /* Inode # */
    dinode     c_node;
    uint8_t    c_refs;            /* In-core reference count */
    bool       c_dirty;           /* Modified flag. */
} cinode, *inoptr;

#define NULLINODE ((inoptr)NULL)
#define NULLINOPTR ((inoptr*)NULL)

#define FILENAME_LENGTH 14
typedef struct direct {
    uint16_t   d_ino;
    char     d_name[14];
} direct;

#define FILESYS_TABSIZE 50
typedef struct filesys { // note: exists in mem and on disk
    int16_t       s_mounted;
    uint16_t      s_isize;
    uint16_t      s_fsize;
    uint16_t      s_nfree;
    blkno_t       s_free[FILESYS_TABSIZE];
    int16_t       s_ninode;
    uint16_t      s_inode[FILESYS_TABSIZE];
    bool          s_fmod; // was int16
    uint8_t       __padding; // s_fmod was previously int16
    uzitime_t        s_time;
    blkno_t       s_tfree;
    uint16_t      s_tinode;
    inoptr        s_mntpt;     /* Mount point */
} filesys, *fsptr;

typedef struct oft {
    off_t     o_ptr;      /* File position pointer */
    inoptr    o_inode;    /* Pointer into in-core inode table */
    uint8_t   o_access;   /* O_RDONLY, O_WRONLY, or O_RDWR */
    uint8_t   o_refs;     /* Reference count: depends on # of active children */
} oft;


/* Process table p_status values */

#define P_EMPTY         0    /* Unused slot */
#define P_RUNNING       1    /* Currently running process (must match value in kernel.def) */
#define P_READY         2    /* Runnable   */
#define P_SLEEP         3    /* Sleeping; can be awakened by signal */
#define P_XSLEEP        4    /* Sleeping, don't wake up for signal */
#define P_PAUSE         5    /* Sleeping for pause(); can wakeup for signal */
#define P_FORKING       6    /* In process of forking; do not mess with */
#define P_WAIT          7    /* Executed a wait() */
#define P_ZOMBIE2       8    /* Exited but code pages still valid. */
#define P_ZOMBIE        9    /* Exited. */


// 0 would also be a valid signal number except u_cursig=0 when no signal is being
// dispatched. workaround would be to use a value > NSIGS to indicate "no signal".
#define SIGHUP  1       /* hangup */
#define SIGINT  2       /* interrupt */
#define SIGQUIT 3       /* quit */
#define SIGILL  4       /* illegal instruction (not reset when caught */
#define SIGTRAP 5       /* trace trap (not reset when caught) */
#define SIGIOT  6       /* IOT instruction */
#define SIGEMT  7       /* EMT instruction */
#define SIGFPE  8       /* floating point exception */
#define SIGKILL 9       /* kill */
#define SIGBUS  10      /* bus error */
#define SIGSEGV 11      /* segmentation violation */
#define SIGSYS  12      /* bad argument to system call */
#define SIGPIPE 13      /* write on a pipe with no one to read it */
#define SIGALRM 14      /* alarm clock */
#define SIGTERM 15      /* software termination signal from kill */

#define  SIG_DFL   (int (*)())0
#define  SIG_IGN   (int (*)())1

#define sigmask(sig)    (1<<(sig))

/* Process table entry */

typedef struct p_tab {
    /* WRS: UPDATE kernel.def IF YOU CHANGE THIS STRUCTURE */
    uint8_t     p_status;       /* Process status: MUST BE FIRST MEMBER OF STRUCT */
    uint8_t     p_tty;          /* Process' controlling tty minor # */
    uint16_t    p_pid;          /* Process ID */
    uint16_t    p_uid;
    struct p_tab *p_pptr;      /* Process parent's table entry */
    uint16_t    p_alarm;        /* Seconds until alarm goes off */
    uint16_t    p_exitval;      /* Exit value */
    void *      p_wait;         /* Address of thing waited for */
    /* Everything below here is overlaid by time info at exit */
    uint16_t    p_page;         /* Process' MMU Base Address */
    uint16_t    p_priority;     /* Process priority */
    uint16_t    p_pending;      /* Bitmask of pending signals */
    uint16_t    p_ignored;      /* Bitmask of ignored signals */
    struct u_block *p_ublk;   /* Pointer to udata block when not running */
    uint16_t    p_waitno;       /* wait #; for finding longest waiting proc */
  /* Added for correct execve() of type 1 binaries */
    uint16_t    p_fork_inf;     /* 0 = not child of fork */
    /* WRS: UPDATE kernel.def IF YOU CHANGE THIS STRUCTURE */

/**HP**/
    char    p_name[8];
    uzitime_t  p_time, p_utime, p_stime, p_cutime, p_cstime;
/**HP**/
} p_tab, *ptptr;

typedef struct u_data {
    /**** If you change this top section, also update offsets in "kernel.def" ****/
    struct p_tab *u_ptab;       /* Process table pointer */
    uint16_t    u_page;         /* Process' MMU Base Address (equal to u_ptab->p_page) */
    bool        u_insys;        /* True if in kernel */
    uint8_t     u_callno;       /* sys call being executed. */
    void *      u_syscall_sp;   /* Stores SP when process makes system call */
    int16_t     u_retval;       /* Return value from sys call */
    int16_t     u_error;        /* Last error number */
    void *      u_sp;           /* Stores SP when process is swapped */
    bool        u_ininterrupt;  /* True when the interrupt handler is runnign (prevents recursive interrupts) */
    int8_t      u_cursig;       /* Next signal to be dispatched */
    uint16_t    u_argn;         /* Last system call arg */
    uint16_t    u_argn1;        /* This way because args on stack backwards */
    uint16_t    u_argn2;
    uint16_t    u_argn3;        /* args n-3, n-2, n-1, and n */
    int16_t     (*u_sigvec[NSIGS])();   /* Array of signal vectors */
    void *      u_isp;          /* Value of initial sp (argv) */
    /**** If you change this top section, also update offsets in "kernel.def" ****/

    char *      u_base;         /* Source or dest for I/O */
    uint16_t    u_count;        /* Amount for I/O */
    off_t       u_offset;       /* Place in file for I/O */
    struct blkbuf *u_buf;
    bool        u_sysio;        /* True if I/O to system space */

    uint16_t    u_gid;
    uint16_t    u_euid;
    uint16_t    u_egid;
    uint16_t    u_mask;         /* umask: file creation mode mask */
    uzitime_t   u_time;         /* Start time */
    uint8_t     u_files[UFTSIZE];       /* Process file table: indices into open file table, or NO_FILE. */
    inoptr      u_cwd;          /* Index into inode table of cwd. */
    uint16_t    u_break;        /* Top of data space */
    //inoptr      u_ino;          /* Used during execve() */
    char        u_name[8];      /* Name invoked with */
    uzitime_t   u_utime;        /* Elapsed ticks in user mode */
    uzitime_t   u_stime;        /* Ticks in system mode */
    uzitime_t   u_cutime;       /* Total childrens ticks */
    uzitime_t   u_cstime;
} u_data;


/* This is the user data structure, padded out to 512 bytes with the
 * System Stack.
 */
typedef struct u_block {
        u_data u_d;
        char   u_s [512 - sizeof(struct u_data)];
} u_block;


/* Struct to temporarily hold arguments in execve */
struct s_argblk {
    int a_argc;
    int a_arglen;
    int a_envc;
    char a_buf[512-3*sizeof(int)];
};

/* Open() parameters. */

#define O_RDONLY        0
#define O_WRONLY        1
#define O_RDWR          2

/*
 * Error codes
 */
#define EPERM           1               /* Not owner */
#define ENOENT          2               /* No such file or directory */
#define ESRCH           3               /* No such process */
#define EINTR           4               /* Interrupted System Call */
#define EIO             5               /* I/O Error */
#define ENXIO           6               /* No such device or address */
#define E2BIG           7               /* Arg list too long */
#define ENOEXEC         8               /* Exec format error */
#define EBADF           9               /* Bad file number */
#define ECHILD          10              /* No children */
#define EAGAIN          11              /* No more processes */
#define ENOMEM          12              /* Not enough core */
#define EACCES          13              /* Permission denied */
#define EFAULT          14              /* Bad address */
#define ENOTBLK         15              /* Block device required */
#define EBUSY           16              /* Mount device busy */
#define EEXIST          17              /* File exists */
#define EXDEV           18              /* Cross-device link */
#define ENODEV          19              /* No such device */
#define ENOTDIR         20              /* Not a directory */
#define EISDIR          21              /* Is a directory */
#define EINVAL          22              /* Invalid argument */
#define ENFILE          23              /* File table overflow */
#define EMFILE          24              /* Too many open files */
#define ENOTTY          25              /* Not a typewriter */
#define ETXTBSY         26              /* Text file busy */
#define EFBIG           27              /* File too large */
#define ENOSPC          28              /* No space left on device */
#define ESPIPE          29              /* Illegal seek */
#define EROFS           30              /* Read-only file system */
#define EMLINK          31              /* Too many links */
#define EPIPE           32              /* Broken pipe */

/* math software */
#define EDOM            33              /* Argument too large */
#define ERANGE          34              /* Result too large */

#define ENAMETOOLONG    63              /* File name too long */

// compiler provides optimised versions of these:
#if defined(__SDCC_z80) || defined(__SDCC_z180) || defined(__SDCC_r2k) || defined(__SDCC_r3ka)
#define memcpy(dst, src, n) __builtin_memcpy(dst, src, n)
#define strcpy(dst, src) __builtin_strcpy(dst, src)
#define strncpy(dst, src, n) __builtin_strncpy(dst, src, n)
#define strchr(s, c) __builtin_strchr(s, c)
#define memset(dst, c, n) __builtin_memset(dst, c, n)
#endif

// functions in common memory
void tempstack(void); // swaps to a temporary stack in common memory

// debug functions
void trap_monitor(void);
void idump(void);

// start.c
void panic(char *deathcry);
bool valaddr(char *base, uint16_t size);

// kdata.c
bool validdev(uint8_t dev);

// usermem.s
uint16_t uget (void *userspace_source, void *dest, uint16_t count);
uint8_t  ugetc(void *userspace_source);
uint16_t ugetw(void *userspace_source);
uint16_t ugets(void *userspace_source, void *dest, uint16_t maxlen);

void uput (void *source,   void *userspace_dest, uint16_t count);
void uputc(uint8_t value,  void *userspace_dest);
void uputw(uint16_t value, void *userspace_dest);
void uzero(void *userspace_dest, uint16_t count);

// tricks.s
void swapout(void);
void doexec(void *start_addr);
void swapin(ptptr process);
int16_t dofork(ptptr child);

// devio.c
char *bread (uint8_t dev, blkno_t blk, bool rewrite);
void brelse(void *bp);
void bawrite(void *bp);
int bfree(bufptr bp, uint8_t dirty); /* dirty: 0=clean, 1=dirty (write back), 2=dirty+immediate write */
void *tmpbuf(void);
void *zerobuf(void);
void bufsync(void);
bufptr bfind(uint8_t dev, blkno_t blk);
bufptr freebuf(void);
void bufinit(void);
void bufdump (void);
int bdread(bufptr bp);
int bdwrite(bufptr bp);
int cdread(uint8_t dev);
int d_open(uint8_t dev);
int d_close(uint8_t dev);
int d_ioctl(uint8_t dev, uint16_t request, char *data);
int cdwrite(uint8_t dev);
bool insq(struct s_queue *q, char c);
bool remq(struct s_queue *q, char *cp);
void clrq(struct s_queue *q);
bool uninsq(struct s_queue *q, char *cp);

// filesys.c
inoptr n_open(char *uname, inoptr *parent); // open file, "name" in user address space
inoptr kn_open(char *uname, inoptr *parent); // open file, "name" in kernel address space
inoptr i_open(uint8_t dev, uint16_t ino);
inoptr srch_dir(inoptr wd, char *compname);
inoptr srch_mt(inoptr ino);
bool ch_link(inoptr wd, char *oldname, char *newname, inoptr nindex);
void filename(char *userspace_upath, char *name);
bool namecomp(char *n1, char *n2); // return true if n1 == n2
inoptr newfile(inoptr pino, char *name);
fsptr getdev(uint8_t dev);
bool baddev(fsptr dev);
uint16_t i_alloc(uint8_t devno);
void i_free(uint8_t devno, uint16_t ino);
blkno_t blk_alloc(uint8_t devno);
void blk_free(uint8_t devno, blkno_t blk);
int8_t oft_alloc(void);
void oft_deref(int8_t of);
int8_t uf_alloc(void); // returns index of slot, or -1 on failure
int8_t uf_alloc(void); // returns index of slot, or -1 on failure
void i_ref(inoptr ino);
void i_deref(inoptr ino);
void wr_inode(inoptr ino);
bool isdevice(inoptr ino);
void f_trunc(inoptr ino);
void freeblk(uint8_t dev, blkno_t blk, uint8_t level);
blkno_t bmap(inoptr ip, blkno_t bn, int rwflg);
void validblk(uint8_t dev, blkno_t num);
inoptr getinode(uint8_t uindex);
bool super(void);
uint8_t getperm(inoptr ino);
void setftime(inoptr ino, uint8_t flag);
uint16_t getmode(inoptr ino);
bool fmount(uint8_t dev, inoptr ino); // returns true on failure, false on success
void magic(inoptr ino);

// inode.c
void readi(inoptr ino);
void writei(inoptr ino);
int16_t doclose (uint8_t uindex);
inoptr rwsetup (bool is_read);

// process.c
void psleep(void *event);
void wakeup(void *event);
ptptr getproc(void);
void newproc(ptptr p);
ptptr ptab_alloc(void);
void ssig(ptptr proc, uint8_t sig);
void chksigs(void);
void program_vectors(uint16_t proc_basepage);
void sgrpsig(uint8_t tty, uint8_t sig);
void unix_syscall(void);
void timer_interrupt(void);
void doexit (int16_t val, int16_t val2);

// syscalls_fs.c, syscalls_proc.c
void updoff(void);
void stcpy(inoptr ino, char *buf);
bool rargs (char **userspace_argv, struct s_argblk *argbuf);
char **wargs(char *userspace_ptr, struct s_argblk *argbuf, int  *cnt);
int16_t __exit(void);        // UZI system call 0
int16_t _open(void);         // UZI system call 1
int16_t _close(void);        // UZI system call 2
int16_t _creat(void);        // UZI system call 3
int16_t _mknod(void);        // UZI system call 4
int16_t _link(void);         // UZI system call 5
int16_t _unlink(void);       // UZI system call 6
int16_t _read(void);         // UZI system call 7
int16_t _write(void);        // UZI system call 8
int16_t _seek(void);         // UZI system call 9
int16_t _chdir(void);        // UZI system call 10
int16_t _sync(void);         // UZI system call 11
int16_t _access(void);       // UZI system call 12
int16_t _chmod(void);        // UZI system call 13
int16_t _chown(void);        // UZI system call 14
int16_t _stat(void);         // UZI system call 15
int16_t _fstat(void);        // UZI system call 16
int16_t _dup(void);          // UZI system call 17
int16_t _getpid(void);       // UZI system call 18
int16_t _getppid(void);      // UZI system call 19
int16_t _getuid(void);       // UZI system call 20
int16_t _umask(void);        // UZI system call 21
int16_t _getfsys(void);      // UZI system call 22
int16_t _execve(void);       // UZI system call 23
int16_t _wait(void);         // UZI system call 24
int16_t _setuid(void);       // UZI system call 25
int16_t _setgid(void);       // UZI system call 26
int16_t _time(void);         // UZI system call 27
int16_t _stime(void);        // UZI system call 28
int16_t _ioctl(void);        // UZI system call 29
int16_t _brk(void);          // UZI system call 30
int16_t _sbrk(void);         // UZI system call 31
int16_t _fork(void);         // UZI system call 32
int16_t _mount(void);        // UZI system call 33
int16_t _umount(void);       // UZI system call 34
int16_t _signal(void);       // UZI system call 35
int16_t _dup2(void);         // UZI system call 36
int16_t _pause(void);        // UZI system call 37
int16_t _alarm(void);        // UZI system call 38
int16_t _kill(void);         // UZI system call 39
int16_t _pipe(void);         // UZI system call 40
int16_t _getgid(void);       // UZI system call 41
int16_t _times(void);        // UZI system call 42
int16_t _utime(void);        // UZI system call 43
int16_t _geteuid(void);      // UZI system call 44
int16_t _getegid(void);      // UZI system call 45

#endif // __UZI__KERNEL_DOT_H__
