#include <version.h>
#include <kernel.h>
#include <kdata.h>
#include <devsd.h>
#include <devrd.h>
#include <devmem.h>
#include <devzero.h>
#include <devnull.h>
#include <devproc.h>
#include <devlpr.h>
#include <devtty.h>

uint16_t ramsize, procmem, maxproc;
uint16_t runticks;
uint16_t system_tick_counter;
bool inint;
uint8_t root_dev;
uint8_t ticks_this_second;
inoptr root;
uint16_t waitno;
uzitime_t tod;    // time of day
uzitime_t ticks;  // system tick counter

struct blkbuf bufpool[NBUFS];   // disk buffer pool
struct p_tab ptab[PTABSIZE];    // process table
struct oft of_tab[OFTSIZE];     // open file handle table
struct cinode i_tab[ITABSIZE];  // open inode table
struct filesys fs_tab[NDEVS];   // filesystem table (TODO: way bigger than it needs to be!!)

#define no_open no_close
int no_close(uint8_t minor)
{
    minor;
    return 0; // OK
}

int no_read(uint8_t minor, uint8_t rawflag)
{
    minor;
    rawflag;
    return -1; // FAIL
}

int no_ioctl(uint8_t minor, uint16_t a, char *b)
{
    minor;
    a;
    b;
    return -1; // FAIL
}

struct devsw dev_tab[] =  /* The device driver switch table */
{
// minor    open         close        read      write       ioctl
// -----------------------------------------------------------------
  /* RAM disk block devices (2MB each, stored in top of RAM */
  {  0,  rd_open,     no_close,    rd_read,   rd_write,   no_ioctl },   //   0   /dev/rd0
  {  1,  rd_open,     no_close,    rd_read,   rd_write,   no_ioctl },   //   1   /dev/rd1
  {  2,  rd_open,     no_close,    rd_read,   rd_write,   no_ioctl },   //   2   /dev/rd2
  {  3,  rd_open,     no_close,    rd_read,   rd_write,   no_ioctl },   //   3   /dev/rd3

  /* SD card block devices (32MB each, stored in type 0x5A partition */
  {  0,  sd_open,     no_close,    sd_read,   sd_write,   no_ioctl },   //   4   /dev/sd0 
  {  1,  sd_open,     no_close,    sd_read,   sd_write,   no_ioctl },   //   5   /dev/sd1 
  {  2,  sd_open,     no_close,    sd_read,   sd_write,   no_ioctl },   //   6   /dev/sd2 
  {  3,  sd_open,     no_close,    sd_read,   sd_write,   no_ioctl },   //   7   /dev/sd3 
  {  4,  sd_open,     no_close,    sd_read,   sd_write,   no_ioctl },   //   8   /dev/sd4 
  {  5,  sd_open,     no_close,    sd_read,   sd_write,   no_ioctl },   //   9   /dev/sd5 
  {  6,  sd_open,     no_close,    sd_read,   sd_write,   no_ioctl },   //  10   /dev/sd6 
  {  7,  sd_open,     no_close,    sd_read,   sd_write,   no_ioctl },   //  11   /dev/sd7 
  {  8,  sd_open,     no_close,    sd_read,   sd_write,   no_ioctl },   //  12   /dev/sd8 
  {  9,  sd_open,     no_close,    sd_read,   sd_write,   no_ioctl },   //  13   /dev/sd9 
  { 10,  sd_open,     no_close,    sd_read,   sd_write,   no_ioctl },   //  14   /dev/sd10
  { 11,  sd_open,     no_close,    sd_read,   sd_write,   no_ioctl },   //  15   /dev/sd11
  { 12,  sd_open,     no_close,    sd_read,   sd_write,   no_ioctl },   //  16   /dev/sd12
  { 13,  sd_open,     no_close,    sd_read,   sd_write,   no_ioctl },   //  17   /dev/sd13
  { 14,  sd_open,     no_close,    sd_read,   sd_write,   no_ioctl },   //  18   /dev/sd14
  { 15,  sd_open,     no_close,    sd_read,   sd_write,   no_ioctl },   //  19   /dev/sd15
  { 16,  sd_open,     no_close,    sd_read,   sd_write,   no_ioctl },   //  20   /dev/sd16
  { 17,  sd_open,     no_close,    sd_read,   sd_write,   no_ioctl },   //  21   /dev/sd17
  { 18,  sd_open,     no_close,    sd_read,   sd_write,   no_ioctl },   //  22   /dev/sd18
  { 19,  sd_open,     no_close,    sd_read,   sd_write,   no_ioctl },   //  23   /dev/sd19
  { 20,  sd_open,     no_close,    sd_read,   sd_write,   no_ioctl },   //  24   /dev/sd20
  { 21,  sd_open,     no_close,    sd_read,   sd_write,   no_ioctl },   //  25   /dev/sd21
  { 22,  sd_open,     no_close,    sd_read,   sd_write,   no_ioctl },   //  26   /dev/sd22
  { 23,  sd_open,     no_close,    sd_read,   sd_write,   no_ioctl },   //  27   /dev/sd23
  { 24,  sd_open,     no_close,    sd_read,   sd_write,   no_ioctl },   //  28   /dev/sd24
  { 25,  sd_open,     no_close,    sd_read,   sd_write,   no_ioctl },   //  29   /dev/sd25
  { 26,  sd_open,     no_close,    sd_read,   sd_write,   no_ioctl },   //  30   /dev/sd26
  { 27,  sd_open,     no_close,    sd_read,   sd_write,   no_ioctl },   //  31   /dev/sd27
  { 28,  sd_open,     no_close,    sd_read,   sd_write,   no_ioctl },   //  32   /dev/sd28
  { 29,  sd_open,     no_close,    sd_read,   sd_write,   no_ioctl },   //  33   /dev/sd29
  { 30,  sd_open,     no_close,    sd_read,   sd_write,   no_ioctl },   //  34   /dev/sd30
  { 31,  sd_open,     no_close,    sd_read,   sd_write,   no_ioctl },   //  35   /dev/sd31

  /* devices below here are not mountable (as per NDEVS) */
  {  0, lpr_open,     lpr_close,   no_read,   lpr_write,  no_ioctl  },  //  36   /dev/lp  
  {  1, tty_open,     tty_close,   tty_read,  tty_write,  tty_ioctl },  //  37   /dev/tty1
  {  2, tty_open,     tty_close,   tty_read,  tty_write,  tty_ioctl },  //  38   /dev/tty2
  {  0, no_open,      no_close,    null_read, null_write, no_ioctl  },  //  39   /dev/null
  {  0, no_open,      no_close,    zero_read, zero_write, no_ioctl  },  //  40   /dev/zero
  {  0, no_open,      no_close,    mem_read,  mem_write,  no_ioctl  },  //  41   /dev/kmem
  {  0, no_open,      no_close,    proc_read, proc_write, proc_ioctl}   //  42   /dev/proc
  // {  0,  mt_open,  mt_close, mt_read,  mt_write,   nogood }    // 43 /dev/mt  
  /* Add more tty channels here if available, incrementing minor# */
};

bool validdev(uint8_t dev)
{
    if(dev >= (sizeof(dev_tab)/sizeof(struct devsw)))
        return false;
    else
        return true;
}

// system call dispatch table
syscall_t syscall_dispatch[UZI_SYSCALL_COUNT] = {
    __exit,        // UZI system call 0
    _open,         // UZI system call 1
    _close,        // UZI system call 2
    _creat,        // UZI system call 3
    _mknod,        // UZI system call 4
    _link,         // UZI system call 5
    _unlink,       // UZI system call 6
    _read,         // UZI system call 7
    _write,        // UZI system call 8
    _seek,         // UZI system call 9
    _chdir,        // UZI system call 10
    _sync,         // UZI system call 11
    _access,       // UZI system call 12
    _chmod,        // UZI system call 13
    _chown,        // UZI system call 14
    _stat,         // UZI system call 15
    _fstat,        // UZI system call 16
    _dup,          // UZI system call 17
    _getpid,       // UZI system call 18
    _getppid,      // UZI system call 19
    _getuid,       // UZI system call 20
    _umask,        // UZI system call 21
    _getfsys,      // UZI system call 22
    _execve,       // UZI system call 23
    _wait,         // UZI system call 24
    _setuid,       // UZI system call 25
    _setgid,       // UZI system call 26
    _time,         // UZI system call 27
    _stime,        // UZI system call 28
    _ioctl,        // UZI system call 29
    _brk,          // UZI system call 30
    _sbrk,         // UZI system call 31
    _fork,         // UZI system call 32
    _mount,        // UZI system call 33
    _umount,       // UZI system call 34
    _signal,       // UZI system call 35
    _dup2,         // UZI system call 36
    _pause,        // UZI system call 37
    _alarm,        // UZI system call 38
    _kill,         // UZI system call 39
    _pipe,         // UZI system call 40
    _getgid,       // UZI system call 41
    _times,        // UZI system call 42
    _utime,        // UZI system call 43
    _geteuid,      // UZI system call 44
    _getegid       // UZI system call 45
};

