.area _LOADER (ABS)

GPIO_OUT = 0x21

; note: cannot use the stack!! SP can be used as a temporary register
; note: cannot use any absolute addresses!

; CP/M will load us at 0x100
; We do not use any absolute addresses so we can safely relocate our code into low memory.
; We copy the entire program to address 0 but we could actually start copying at next_block,
; however this would require us to convince the assembler that next_block is at 0.

    .org 0x0000
    ; setup
    di
    ; copy ourselves to the very bottom of memory -- 0x00 upwards
    ld de, #0
    ld hl, #0x100                    ; start of code to copy, offset by CP/M load address
    ld bc, #(datastart)              ; length of our code
    ldir                             ; copy copy copy!
    ld hl, #(datastart + 0x100)      ; start of data trailer, offset by CP/M load address
    ; jump to next_block in the new location in low memory
    jp next_block   ; this is the only absolute jump we perform
    ; process a block from the data stream
next_block:
    ld a, (hl)
    inc hl
    cp #0x00 ; load block?
    jr z, do_load_forward
;    cp #0x01 ; load backwards?
;    jr z, do_load_backwards
    cp #0x02 ; fill zero?
    jr z, do_fill_zero
    cp #0xfe ; skip?
    jr z, do_skip
    cp #0xff ; jump?
    jr z, do_jump
    ; hmmm, if we get here, something went wrong.
failed:
    ; turn LEDs in 1010 pattern
    ld a, #0x0a
    out (GPIO_OUT), a
    halt ; zzzzzzzz
    jr failed

do_load_backwards:
do_load_forward:
    ; load block (type 0x00), format: 16 bit target address (first address to write), 16 bit length, data (length bytes)
    ; load block (type 0x01), format: 16 bit target address (LAST address to write), 16 bit length, data (length bytes)
    ld e, (hl)
    inc hl
    ld d, (hl)
    inc hl
    ld c, (hl)
    inc hl
    ld b, (hl)
    inc hl
    ; now A=block type, DE=destination, HL=source, BC=count
;    cp #0x00 ; was it a forward load?
;    jr nz, backwards
    ldir
    jr next_block
;backwards:
;    ; setup to copy backwards
;    add hl, bc ; HL now points at the first byte of the NEXT block (ie one byte past the end of our data)
;    ; use SP to store a copy of HL
;    ld sp, hl
;    ; point HL at the last byte of our data
;    dec hl
;    ; do the copy
;    lddr
;    ; point HL at the next block (address was stashed in SP)
    jr restorehl

do_fill_zero:
    ; fill with zero (type 0x02), format: 16 bit target address, 16 bit length
    ld e, (hl)
    inc hl
    ld d, (hl)
    inc hl
    ld c, (hl)
    inc hl
    ld b, (hl)
    inc hl
    ; use SP to store a copy HL
    ld sp, hl
    ; setup to fill with zeroes
    ld hl, #0
    add hl, de
    ld (hl), #0
    inc de
    dec bc
    ld a, b ; just check length>0 (length=1 would be valid)
    or c
    jr z, restorehl
    ldir
    ; restore HL to point at the next block, then process it
restorehl:
    ld hl, #0
    add hl, sp
    jr next_block

do_skip:
    ld e, (hl)
    inc hl
    ld d, (hl)
    inc hl
    add hl, de
    jr next_block

do_jump:
    ; jump block (type 0xff), format: 16 bit target address
    ld e, (hl)
    inc hl
    ld d, (hl)
    inc hl
    ex de, hl
    jp (hl)
    ; we never return
    
    ; the data is in a trailer
    ; note that CP/M loads this offset by 0x100 bytes
datastart:
    ; data is laid out in blocks.
    ; first byte of block is the block type:
    ;   0x00   load block (forwards)
    ;          - then 16-bit target address (FIRST address to write), 16-bit length, data (length bytes)
    ;   0x01   load block (backwards)
    ;          - then 16-bit target address (LAST address to write), 16-bit length, data (length bytes)
    ;   0x02   fill zeros
    ;          - then 16-bit target address (FIRST address to write), 16-bit length
    ;   0xFE   skip
    ;          - then 16-bit count of bytes to skip ahead, the following count bytes are ignored.
    ;   0xFF   jump to address
    ;          - payload is two bytes, address to jump to
    ;   The final block must be type 0xFF.
