; 2013-12-21 William R Sowerbutts

        .module syscall

        ; exported symbols
        .globl unix_syscall_entry

        ; imported symbols
        .globl _trap_monitor
        .globl _unix_syscall
        .globl outstring
        .globl kstack_top
        .globl mmu_map_process
        .globl dispatch_process_signal

        .include "kernel.def"

        .area _COMMONMEM

; entry point for UZI system calls
unix_syscall_entry:
        di
        ; store processor state
        ex af, af'
        push af
        ex af, af'
        exx
        push bc
        push de
        push hl
        exx
        ; push af ;; WRS: also skip this
        push bc
        push de
        ; push hl ;; WRS: we could skip this since we always set HL on return
        push ix
        push iy

        ; locate function call arguments on the userspace stack
        ld hl, #18     ; 16 bytes machine state, plus 2 bytes return address
        add hl, sp
        ; save system call number
        ld a, (hl)
        ld (U_DATA__U_CALLNO), a
        ; advance to syscall arguments
        inc hl
        inc hl
        ; copy arguments to common memory
        ld bc, #8      ; four 16-bit values
        ld de, #U_DATA__U_ARGN
        ldir           ; copy

        ; save process stack pointer
        ld (U_DATA__U_SYSCALL_SP), sp
        ; switch to kernel stack
        ld sp, #kstack_top

        ; map in kernel
        ld hl, #OS_BANK
        call mmu_map_process

        ; re-enable interrupts
        ei

        ; now pass control to C
        call _unix_syscall

        di
        ; map process memory back in
        ld hl, (U_DATA__U_PAGE)
        call mmu_map_process

        ; switch back to user stack
        ld sp, (U_DATA__U_SYSCALL_SP)

        ; check for signals, call the handlers
        call dispatch_process_signal

        ; check if error condition to be signalled on return
        ld hl, (U_DATA__U_ERROR)
        ld a, h
        or l    ; set NZ flag if we are to return error
        jr z, not_error

        scf    ; set carry flag
        ; note error code remains in HL
        jr unix_return

not_error:
        ; no error to signal! return syscall return value instead of error code
        ld hl, (U_DATA__U_RETVAL)
        ; fall through to return code

unix_return:
        ; restore machine state
        pop iy
        pop ix
        ; pop hl ;; WRS: skip this!
        pop de
        pop bc
        ; pop af ;; WRS: skip this!
        exx
        pop hl
        pop de
        pop bc
        exx
        ex af, af'
        pop af
        ex af, af'
        ei
        ret ; must immediately follow EI
