#include <kernel.h>
#include <kdata.h>
#include <printf.h>
#include <timer.h>

void updoff(void)
{
    off_t *offp;

    /* Update current file pointer */
    offp = &of_tab[udata.u_files[udata.u_argn]].o_ptr;
    offp->o_blkno = udata.u_offset.o_blkno;
    offp->o_offset = udata.u_offset.o_offset;
}


/*******************************************
  seek (file, offset, flag)         Function 9        ?
  int16_t file;
  uint16_t offset;
  int16_t flag;
 ********************************************/
#define file   (uint8_t)udata.u_argn
#define offset (uint16_t)udata.u_argn1
#define flag   (int16_t)udata.u_argn2

int16_t _seek(void)
{
    inoptr ino;
    int16_t oftno;
    uint16_t retval;

    if((ino = getinode (file)) == NULLINODE)
        return (-1);

    if(getmode (ino) == F_PIPE)
    {
        udata.u_error = ESPIPE;
        return (-1);
    }

    oftno = udata.u_files[file];

    /* REMEMBER:
     * flag  0-2 return only offset within current block (512 bytes)
     *       3-5 return only current block number
     */
    if(flag <= 2)
        retval = of_tab[oftno].o_ptr.o_offset;
    else
        retval = of_tab[oftno].o_ptr.o_blkno;

    switch (flag)
    {
        case 0:
            of_tab[oftno].o_ptr.o_blkno = 0;
            of_tab[oftno].o_ptr.o_offset = offset;
            break;
        case 1:
            of_tab[oftno].o_ptr.o_offset += offset;
            break;
        case 2:
            of_tab[oftno].o_ptr.o_blkno = ino->c_node.i_size.o_blkno;
            of_tab[oftno].o_ptr.o_offset = ino->c_node.i_size.o_offset + offset;
            break;
        case 3:
            of_tab[oftno].o_ptr.o_blkno = offset;
            break;
        case 4:
            of_tab[oftno].o_ptr.o_blkno += offset;
            break;
        case 5:
            of_tab[oftno].o_ptr.o_blkno = ino->c_node.i_size.o_blkno + offset;
            break;
        default:
            udata.u_error = EINVAL;
            return (-1);
    }
    while ((unsigned)of_tab[oftno].o_ptr.o_offset >= 512) {
        of_tab[oftno].o_ptr.o_offset -= 512;
        ++of_tab[oftno].o_ptr.o_blkno;
    }

    return ((int16_t)retval);
}

#undef file
#undef offset
#undef flag


/*******************************************
  chdir (dir)                      Function 10
  char *dir;
 ********************************************/
#define dir (char *)udata.u_argn

int16_t _chdir(void)
{
    inoptr newcwd;

    if(! (newcwd = n_open (dir, NULLINOPTR)))
        return (-1);

    if(getmode (newcwd) != F_DIR)
    {
        udata.u_error = ENOTDIR;
        i_deref (newcwd);
        return (-1);
    }
    i_deref (udata.u_cwd);
    udata.u_cwd = newcwd;
    return (0);
}

#undef dir


/*******************************************
  mknod (name, mode, dev)           Function 4
  char  *name;
  int16_t mode;
  int16_t dev;
 ********************************************/
#define name (char *)udata.u_argn
#define mode (int16_t)udata.u_argn1
#define dev  (int16_t)udata.u_argn2

int16_t _mknod(void)
{
    inoptr ino;
    inoptr parent;
    char   fname[15];

    udata.u_error = 0;

    if(ino = n_open (name, &parent)) {
        udata.u_error = EEXIST;
        goto nogood;
    }

    if(!parent) {
        udata.u_error = ENOENT;
        return (-1);
    }

    filename (name, fname);
    if(!(*fname && (ino = newfile (parent, fname))))
        goto nogood3;       /* parent inode is derefed in newfile. SN */

    /* Initialize mode and dev */
    ino->c_node.i_mode = mode & ~udata.u_mask;
    ino->c_node.i_addr[0] = isdevice (ino) ? dev : 0;
    setftime(ino, A_TIME|M_TIME|C_TIME);
    wr_inode(ino);

    i_deref (ino);
    return (0);

nogood:
    i_deref (ino);
    i_deref (parent);
nogood3:
    return (-1);
}

#undef name
#undef mode
#undef dev


/*******************************************
  sync()
 ********************************************/

int16_t _sync(void)
{
    uint8_t j;
    inoptr ino;
    char *buf;

    /* Write out modified inodes */

    for (ino=i_tab; ino < i_tab+ITABSIZE; ++ino)
        if(ino->c_refs > 0 && ino->c_dirty != 0) {
            wr_inode (ino);
            ino->c_dirty = 0;
        }

    /* Write out modified super blocks */
    /* This fills the rest of the super block with garbage. */

    for (j=0; j < NDEVS; ++j) {
        if(fs_tab[j].s_mounted == SMOUNTED && fs_tab[j].s_fmod) {
            fs_tab[j].s_fmod = false;
            buf = bread(j, 1, 1);
            memcpy(buf, (char *)&fs_tab[j], 512);
            bfree((bufptr)buf, 2);
        }
    }

    bufsync();          /* Clear buffer pool */

    return 0;
}


/*******************************************
  access (path, mode)              Function 12        ?
  char  *path;
  int16_t mode;
 ********************************************/
#define path (char *)udata.u_argn
#define mode (int16_t)udata.u_argn1

int16_t _access(void)
{
    inoptr ino;
    uint16_t euid;
    uint16_t egid;
    uint16_t retval;

    if((mode & 07) && !ugetc(path))
    {
        udata.u_error = ENOENT;
        return (-1);
    }

    /* Temporarily make eff. id real id. */
    euid = udata.u_euid;
    egid = udata.u_egid;
    udata.u_euid = udata.u_ptab->p_uid;
    udata.u_egid = udata.u_gid;

    if(!(ino = n_open (path, NULLINOPTR))) {
        retval = -1;
        goto nogood;
    }

    retval = 0;
    if(~getperm (ino) & (mode&07)) {
        udata.u_error = EPERM;
        retval = -1;
    }

    i_deref (ino);
nogood:
    udata.u_euid = euid;
    udata.u_egid = egid;

    return (retval);
}

#undef path
#undef mode


/*******************************************
  chmod (path, mode)               Function 13
  char  *path;
  int16_t mode;
 ********************************************/
#define path (char *)udata.u_argn
#define mode (int16_t)udata.u_argn1

int16_t _chmod(void)
{
    inoptr ino;

    if(!(ino = n_open (path, NULLINOPTR)))
        return (-1);

    if(ino->c_node.i_uid != udata.u_euid && !super())
    {
        i_deref (ino);
        udata.u_error = EPERM;
        return (-1);
    }

    ino->c_node.i_mode = (mode & MODE_MASK) | (ino->c_node.i_mode & F_MASK);
    setftime (ino, C_TIME);
    i_deref (ino);
    return (0);
}

#undef path
#undef mode


/*******************************************
  chown (path, owner, group)       Function 14        ?
  char *path;
  int  owner;
  int  group;
 ********************************************/
#define path (char *)udata.u_argn
#define owner (int16_t)udata.u_argn1
#define group (int16_t)udata.u_argn2

int16_t _chown(void)
{
    inoptr ino;

    if(!(ino = n_open (path, NULLINOPTR)))
        return (-1);

    if(ino->c_node.i_uid != udata.u_euid && !super()) {
        i_deref (ino);
        udata.u_error = EPERM;
        return (-1);
    }
    ino->c_node.i_uid = owner;
    ino->c_node.i_gid = group;
    setftime (ino, C_TIME);
    i_deref (ino);
    return (0);
}

#undef path
#undef owner
#undef group


/*******************************************
  utime (file, buf)                Function 43
  char *file;
  char *buf;
 ********************************************/
#define file (char *)udata.u_argn
#define buf (char *)udata.u_argn1

int16_t _utime(void)
{
    inoptr ino;

    if(!valaddr(buf, 2*sizeof (uzitime_t)))
        return (-1);
    if(!(ino = n_open (file, NULLINOPTR)))
        return (-1);
    if(ino->c_node.i_uid != udata.u_euid && !super())
    {
        i_deref (ino);
        udata.u_error = EPERM;
        return (-1);
    }
    /* If Source is in Common Memory (F000-FFFF is always in context),
     * simply move the data.  Otherwise perform interbank "Get".
     */
    if(buf >= PROGTOP)
        memcpy(&(ino->c_node.i_atime), buf, 2*sizeof (uzitime_t));
    else
        uget(buf, &(ino->c_node.i_atime), 2*sizeof (uzitime_t));
    setftime (ino, C_TIME);
    i_deref (ino);
    return (0);
}

#undef file
#undef buf


/*******************************************
  stat (path, buf)                 Function 15
  char *path;
  char *buf;
 ********************************************/
#define path (char *)udata.u_argn
#define buf (char *)udata.u_argn1

int16_t _stat(void)
{
    inoptr ino;
    if(!(ino = n_open (path, NULLINOPTR)))
        return (-1);
    stcpy (ino, buf);
    i_deref (ino);
    return (0);
}

#undef path
#undef buf


/*******************************************
  fstat (fd, buf)                  Function 16
  int16_t fd;
  char *buf;
 ********************************************/
#define fd (int16_t)udata.u_argn
#define buf (char *)udata.u_argn1

int16_t _fstat(void)
{
    inoptr ino;

    if(!(valaddr (buf, sizeof (struct stat))))
        return (-1);

    if((ino = getinode (fd)) == NULLINODE)
        return (-1);

    stcpy (ino, buf);
    return (0);
}

#undef fd
#undef buf


/* Utility for stat and fstat */
void stcpy (inoptr ino, char *buf)
{
    /*   !!! violently system-dependent !!! */
    /* If Destination is in Common Memory (F000-FFFF is always in context),
     * simply move, otherwise perform Interbank "uput" to User Bank.
     */
    if(buf >= PROGTOP) {
        memcpy(buf, (char *)&(ino->c_dev), 12);
        memcpy(buf+12, (char *)&(ino->c_node.i_addr[0]), 2);
        memcpy(buf+14, (char *)&(ino->c_node.i_size), 16);
    } else {
        uput((char *)&(ino->c_dev), buf, 12);
        uput((char *)&(ino->c_node.i_addr[0]), buf+12, 2);
        uput((char *)&(ino->c_node.i_size), buf+14, 16);
    }
}


/*******************************************
  dup (oldd)                       Function 17
  int16_t oldd;
 ********************************************/
#define oldd (uint16_t)udata.u_argn

int16_t _dup(void)
{
    int8_t newd;

    if(getinode (oldd) == NULLINODE)
        return (-1);

    if((newd = uf_alloc()) == -1)
        return (-1);

    udata.u_files[newd] = udata.u_files[oldd];
    ++of_tab[udata.u_files[oldd]].o_refs;

    return (newd);
}

#undef oldd


/*******************************************
  dup2 (oldd, newd)                Function 36        ?
  int16_t oldd;
  int16_t newd;
 ********************************************/
#define oldd (int16_t)udata.u_argn
#define newd (int16_t)udata.u_argn1

int16_t _dup2(void)
{

    if(getinode (oldd) == NULLINODE)
        return (-1);

    if(newd < 0 || newd >= UFTSIZE) {
        udata.u_error = EBADF;
        return (-1);
    }

    if(udata.u_files[newd] != NO_FILE)
        doclose (newd);

    udata.u_files[newd] = udata.u_files[oldd];
    ++of_tab[udata.u_files[oldd]].o_refs;

    return (0);
}

#undef oldd
#undef newd


/*******************************************
  umask (mask)                     Function 21        ?
  int mask;
 ********************************************/
#define mask (int16_t)udata.u_argn

int16_t _umask(void)
{
    int omask;

    omask = udata.u_mask;
    udata.u_mask = mask & 0777;
    return (omask);
}

#undef mask


/* Special system call returns super-block of given filesystem
 * for users to determine free space, etc.  Should be replaced
 * with a sync() followed by a read of block 1 of the device.
 */
/*******************************************
  getfsys (dev, buf)               Function 22
  int16_t  dev;
  struct filesys *buf;
 ********************************************/
#define dev (uint8_t)udata.u_argn
#define buf (struct filesys *)udata.u_argn1

int16_t _getfsys(void)
{
    if(dev == NO_DEVICE || dev >= NDEVS || fs_tab[dev].s_mounted != SMOUNTED) {
        udata.u_error = ENXIO;
        return (-1);
    }
    if((char*)(buf) >= PROGTOP) /* If in Common Memory, simply move */
        memcpy((char*)buf, (char *)&fs_tab[dev], sizeof (struct filesys));
    else
        uput((char *)&fs_tab[dev], (char *)buf, sizeof (struct filesys));
    return (0);
}

#undef dev
#undef buf


/*******************************************
  ioctl (fd, request, data)        Function 29
  int  fd;
  int  request;
  char *data;
 ********************************************/
#define fd (int)udata.u_argn
#define request (int)udata.u_argn1
#define data (char *)udata.u_argn2

int16_t _ioctl(void)
{
    inoptr ino;
    uint8_t dev;

    if((ino = getinode (fd)) == NULLINODE)
        return (-1);

    if(!(isdevice (ino))) {
        udata.u_error = ENOTTY;
        return (-1);
    }

    if(!(getperm (ino) & OTH_WR)) {
        udata.u_error = EPERM;
        return (-1);
    }

    dev = ino->c_node.i_addr[0];

    if(d_ioctl (dev, request, data))
        return (-1);
    return (0);
}

#undef fd
#undef request
#undef data


/* This implementation of mount ignores the rwflag */
/*******************************************
  mount (spec, dir, rwflag)        Function 33
  char *spec;
  char *dir;
  int  rwflag;
 ********************************************/
#define spec (char *)udata.u_argn
#define dir (char *)udata.u_argn1
#define rwflag (int)udata.u_argn2

int16_t _mount(void)
{
    inoptr sino, dino;
    uint8_t dev;

    if(!super()) {
        udata.u_error = EPERM;
        return (-1);
    }

    if(!(sino = n_open (spec, NULLINOPTR)))
        return (-1);

    if(!(dino = n_open (dir, NULLINOPTR))) {
        i_deref (sino);
        return (-1);
    }

    if(getmode (sino) != F_BDEV) {
        udata.u_error = ENOTBLK;
        goto nogood;
    }

    if(getmode (dino) != F_DIR) {
        udata.u_error = ENOTDIR;
        goto nogood;
    }

    dev = (int)sino->c_node.i_addr[0];

    if( !validdev(dev) || d_open (dev)) {
        udata.u_error = ENXIO;
        goto nogood;
    }

    if(fs_tab[dev].s_mounted == SMOUNTED || dino->c_refs != 1 || dino->c_num == ROOTINODE) {
        udata.u_error = EBUSY;
        goto nogood;
    }

    _sync();

    if(fmount (dev, dino)) {
        udata.u_error = EBUSY;
        goto nogood;
    }

    i_deref (dino);
    i_deref (sino);
    return (0);

nogood:
    i_deref (dino);
    i_deref (sino);
    return (-1);
}

#undef spec
#undef dir
#undef rwflag


/*******************************************
  umount (spec)                    Function 34
  char *spec;
 ********************************************/
#define spec (char *)udata.u_argn

int16_t _umount(void)
{
    inoptr sino;
    uint8_t dev;
    inoptr ptr;

    if(!super()) {
        udata.u_error = EPERM;
        return (-1);
    }

    if(!(sino = n_open (spec, NULLINOPTR)))
        return (-1);

    if(getmode (sino) != F_BDEV) {
        udata.u_error = ENOTBLK;
        goto nogood;
    }

    dev = (int)sino->c_node.i_addr[0];
    if(!validdev(dev)) {
        udata.u_error = ENXIO;
        goto nogood;
    }

    if(!fs_tab[dev].s_mounted) {
        udata.u_error = EINVAL;
        goto nogood;
    }

    for (ptr = i_tab; ptr < i_tab+ITABSIZE; ++ptr)
        if(ptr->c_refs > 0 && ptr->c_dev == dev) {
            udata.u_error = EBUSY;
            goto nogood;
        }

    _sync();

    fs_tab[dev].s_mounted = 0;
    i_deref(fs_tab[dev].s_mntpt);
    i_deref(sino);
    return 0;

nogood:
    i_deref (sino);
    return -1;
}

#undef spec

/*******************************************
open (name, flag)                 Function 1
char *name;
int16_t flag;
********************************************/
#define name (char *)udata.u_argn
#define flag (int16_t)udata.u_argn1

int16_t _open(void)
{
    int8_t uindex;
    int8_t oftindex;
    inoptr ino;
    int16_t perm;

    if(flag < 0 || flag > 2) {
        udata.u_error = EINVAL;
        return (-1);
    }
    if((uindex = uf_alloc()) == -1)
        return (-1);

    if((oftindex = oft_alloc()) == -1)
        goto nooft;

    if(!(ino = n_open (name, NULLINOPTR)))
        goto cantopen;

    of_tab[oftindex].o_inode = ino;

    perm = getperm (ino);
    if(((flag == O_RDONLY || flag == O_RDWR) && !(perm & OTH_RD)) ||
        ((flag == O_WRONLY || flag == O_RDWR) && !(perm & OTH_WR))) {
        udata.u_error = EPERM;
        goto cantopen;
    }

    if(getmode (ino) == F_DIR &&
        (flag == O_WRONLY || flag == O_RDWR)) {
        udata.u_error = EISDIR;
        goto cantopen;
    }

    if(isdevice (ino) && d_open ((int)ino->c_node.i_addr[0]) != 0) {
        udata.u_error = ENXIO;
        goto cantopen;
    }

    udata.u_files[uindex] = oftindex;

    of_tab[oftindex].o_ptr.o_offset = 0;
    of_tab[oftindex].o_ptr.o_blkno = 0;
    of_tab[oftindex].o_access = flag;

/*
 *         Sleep process if no writer or reader
 */
    if(getmode (ino) == F_PIPE && of_tab[oftindex].o_refs == 1 )
        psleep (ino);

    return (uindex);

cantopen:
    oft_deref (oftindex);      /* This will call i_deref() */
nooft:
    udata.u_files[uindex] = NO_FILE;
    return (-1);
}

#undef name
#undef flag



/*******************************************
close (uindex)                    Function 2
int16_t uindex;
********************************************/
#define uindex (int8_t)udata.u_argn

int16_t _close(void)
{
    return (doclose (uindex));
}

#undef uindex



/*******************************************
creat (name, mode)                Function 3
char  *name;
int16_t mode;
********************************************/
#define name (char *)udata.u_argn
#define mode (int16_t)udata.u_argn1

int16_t _creat(void)
{
    inoptr ino;
    int8_t uindex;
    int8_t oftindex;
    inoptr   parent;
    int16_t j;

    char fname[15];

    parent = NULLINODE;

    if((uindex = uf_alloc()) == -1){
        return (-1);
    }
    if((oftindex = oft_alloc()) == -1){
        return (-1);
    }

    if(ino = n_open (name, &parent))  /* The file exists */
    {
        i_deref(parent);

        if(getmode (ino) == F_DIR)
        {
            i_deref (ino);
            udata.u_error = EISDIR;
            goto nogood;
        }
        if(!(getperm (ino) & OTH_WR))
        {
            i_deref (ino);
            udata.u_error = EACCES;
            goto nogood;
        }
        if(getmode (ino) == F_REG)
        {
            /* Truncate the file to zero length */
            f_trunc (ino);
            /* Reset any oft pointers */
            for (j=0; j < OFTSIZE; ++j)
                if(of_tab[j].o_inode == ino)
                    of_tab[j].o_ptr.o_blkno = of_tab[j].o_ptr.o_offset = 0;
        }
    }
    else
    {
        /* Get trailing part of name into fname */
        filename(name, fname);
        if(parent && *fname && (ino = newfile (parent, fname)))
                 /* Parent was derefed in newfile */
        {
            ino->c_node.i_mode = (F_REG | (mode & MODE_MASK & ~udata.u_mask));
            setftime (ino, A_TIME|M_TIME|C_TIME);
            /* The rest of the inode is initialized in newfile() */
            wr_inode (ino);
        }
        else
        {
            /* Doesn't exist and can't make it */
            udata.u_error = ENFILE;
            goto nogood;
        }
    }
    udata.u_files[uindex] = oftindex;

    of_tab[oftindex].o_ptr.o_offset = 0;
    of_tab[oftindex].o_ptr.o_blkno = 0;
    of_tab[oftindex].o_inode = ino;
    of_tab[oftindex].o_access = O_WRONLY;

    return (uindex);

nogood:
    oft_deref (oftindex);
    return (-1);
}

#undef name
#undef mode



/*******************************************
pipe (fildes)                    Function 40		?
int fildes[];
********************************************/
#define fildes (int16_t *)udata.u_argn

int16_t _pipe(void)
{
    int8_t u1, u2, oft1, oft2;
    inoptr ino;

/* bug fix SN */
    if((u1 = uf_alloc()) == -1)
        goto nogood;
    if((oft1 = oft_alloc()) == -1)
        goto nogood;
    udata.u_files[u1] = oft1;

    if((u2 = uf_alloc()) == -1)
        goto nogood2;
    if((oft2 = oft_alloc()) == -1)
        goto nogood2;

    if(!(ino = i_open (root_dev, 0))) {
        oft_deref (oft2);
        goto nogood2;
    }

    udata.u_files[u2] = oft2;

    of_tab[oft1].o_ptr.o_offset = 0;
    of_tab[oft1].o_ptr.o_blkno = 0;
    of_tab[oft1].o_inode = ino;
    of_tab[oft1].o_access = O_RDONLY;

    of_tab[oft2].o_ptr.o_offset = 0;
    of_tab[oft2].o_ptr.o_blkno = 0;
    of_tab[oft2].o_inode = ino;
    of_tab[oft2].o_access = O_WRONLY;

    ++ino->c_refs;
    ino->c_node.i_mode = F_PIPE | 0777; /* No permissions necessary on pipes */
    ino->c_node.i_nlink = 0;            /* a pipe is not in any directory */

    // write results to userspace
    uputw(u1, fildes);
    uputw(u2, fildes+1);
    return (0);

nogood2:
    oft_deref (oft1);
    udata.u_files[u1] = NO_FILE;
nogood:
    return (-1);
}

#undef fildes



/*******************************************
link (name1, name2)               Function 5
char *name1;
char *name2;
********************************************/
#define name1 (char *)udata.u_argn
#define name2 (char *)udata.u_argn1

int16_t _link(void)
{
    inoptr ino;
    inoptr ino2;
    inoptr   parent2;
    char     fname[15];

    if(!(ino = n_open (name1, NULLINOPTR)))
        return (-1);

/*    if(getmode(ino) == F_DIR && !super()) */
    if(!(getperm (ino) & OTH_WR) && !super()) {
        udata.u_error = EPERM;
        goto nogood;
    }

      /* Make sure file2 doesn't exist, and get its parent */
    if(ino2 = n_open (name2, &parent2)) {
        i_deref (ino2);
        i_deref (parent2);
        udata.u_error = EEXIST;
        goto nogood;
    }

    if(!parent2)
        goto nogood;

    if(ino->c_dev != parent2->c_dev) {
        i_deref (parent2);
        udata.u_error = EXDEV;
        goto nogood;
    }

    filename(name2, fname);

    if(!(*fname && ch_link (parent2, "", fname, ino))) {
        i_deref (parent2);
        goto nogood;
    }

    /* Update the link count. */
    ++ino->c_node.i_nlink;
    wr_inode(ino);
    setftime(ino, C_TIME);

    i_deref(parent2);
    i_deref(ino);
    return 0;

nogood:
    i_deref (ino);
    return -1;
}

#undef name1
#undef name2



/*******************************************
unlink (path)                     Function 6
char *path;
********************************************/
#define path (char *)udata.u_argn

int16_t _unlink(void)
{
    inoptr ino;
    inoptr   pino;
    char     fname[15];

    ino = n_open (path, &pino);

/* BUG !!
 *  resulting in too many i-refs to pino for non existing files.
 * Fixed SN */
    if(!(pino && ino)) {
/* FIX ... */
        if(pino)       /* parent exist */
            i_deref (pino);
/* ... FIX end */
        udata.u_error = ENOENT;
        return (-1);
    }

/*    if(getmode (ino) == F_DIR && !super()) */

      /* Owner of file can overwrite R/O mode
       * ROOT has no restrictions
       */
    if(ino->c_node.i_uid != udata.u_euid  &&
        !(getperm (ino) & OTH_WR)  &&  !super()) {
        udata.u_error = EPERM;
        goto nogood;
    }

      /* Remove the directory entry */
    filename (path, fname);
    if(!(*fname && ch_link (pino, fname, "", NULLINODE)))
        goto nogood;

      /* Decrease the link count of the inode */
    if(!(ino->c_node.i_nlink--)) {
        ino->c_node.i_nlink += 2;
        kprintf("_unlink: bad nlink\n");
    }
    setftime (ino, C_TIME);
    i_deref (pino);
    i_deref (ino);
    return (0);

nogood:
    i_deref (pino);
    i_deref (ino);
    return (-1);
}

#undef path



/*******************************************
read (d, buf, nbytes)             Function 7
int16_t  d;
char   *buf;
uint16_t nbytes;
********************************************/
#define d (int16_t)udata.u_argn
#define buf (char *)udata.u_argn1
#define nbytes (uint16_t)udata.u_argn2

int16_t _read(void)
{
    inoptr ino;

    /* Set up u_base, u_offset, ino; check permissions, file num. */
    if((ino = rwsetup(true)) == NULLINODE)
        return (-1);          /* bomb out if error */

    readi (ino);
    updoff();

    return (udata.u_count);
}

#undef d
#undef buf
#undef nbytes


/*******************************************
write (d, buf, nbytes)            Function 8
int16_t  d;
char   *buf;
uint16_t nbytes;
********************************************/
#define d (int16_t)udata.u_argn
#define buf (char *)udata.u_argn1
#define nbytes (uint16_t)udata.u_argn2

int16_t _write(void)
{
    inoptr ino;

    /* Set up u_base, u_offset, ino; check permissions, file num. */
    if((ino = rwsetup(false)) == NULLINODE)
        return (-1);          /* bomb out if error */

    writei(ino);
    updoff();

    return (udata.u_count);
}

#undef d
#undef buf
#undef nbytes

