#include <kernel.h>
#include <kdata.h>
#include <timer.h>
#include <printf.h>

/* the interrupt handler increments this after every timer interrupt */
uint16_t system_timer;

/* WRS: timer functions */
timer_t set_timer_duration(uint16_t duration)
{
    if(duration & 0x8000){
        kprintf("bad timer duration 0x%x\n", duration);
    }
    return (system_timer+duration);
}

bool timer_expired(timer_t timer_val)
{
    if((timer_val - system_timer) > 0x7fff)
        return true;
    return false;
}



/* This adds two tick counts together.  The t_time field holds up to
 *  one second of ticks, while the t_date field counts minutes.
 */

void addtick(uzitime_t *t1, uzitime_t *t2)
{
    t1->t_time += t2->t_time;
    t1->t_date += t2->t_date;
    if (t1->t_time >= 60*TICKSPERSEC)
    {
        t1->t_time -= 60*TICKSPERSEC;
        ++t1->t_date;
    }
}


void incrtick (uzitime_t *t)
{
    if (++t->t_time == 60*TICKSPERSEC)
    {
        t->t_time = 0;
        ++t->t_date;
    }
}

/*-----------------------------------------------------------*/
/* Read date/time to System location from global variable    */
void rdtime (uzitime_t *tloc)
{
    __critical{
        tloc->t_date = tod.t_date;
        tloc->t_time = tod.t_time;
    }
}
